#include "header.h"
#include "blackScholes.h"
#include "util.h"
#include <math.h>
#include <stdio.h>

/* Calculates the mesh estimae of [t][p] using inner control 3.
 * Updates Prob[t][p][u].  Prob[t][p][u] should be initialized before
 * the first call.  */
double mesh_estimate_3(int t, int p){
  int p2, points, u;
  double S1,S2,S7,S8,S9,S10,S12,S13,g[POINTS_SIZE],G,beta,S7U[UNDERLIERS_SIZE],S8U[UNDERLIERS_SIZE];
  int count_g;

  weights1(t,p,t+1);
  points = points_on_timeslice(t+1);

  G = exp(-INTEREST*t)*maxOf2Risk(Price[t][ Max1[t][p] ][p], Price[t][ Max2[t][p] ][p], STRIKE, 1.0, sqrt(VARIANCE),
				  sqrt(VARIANCE), INTEREST, DIVIDEND, DIVIDEND, 0.0);

  for(p2=0;p2<points;++p2){
    g[p2] = dMax(Price[t+1][ Max1[t][p] ][p2], Price[t+1][ Max2[t][p] ][p2]);
    g[p2]  = dMax(0, exp(-INTEREST*(t+1))*(g[p2] - STRIKE));
  }

  S1 = 0;
  S2 = 0;
  S7 = 0;
  S8 = 0;
  S9 = 0;
  S10 = 0;
  S12 = 0;
  S13 = 0;

  for(u=0;u<UNDERLIERS;++u){
    S7U[u] = 0.0;
    S8U[u] = 0.0;
  }


  count_g = 0;
  for (p2 = 0;p2<UNDERLIERS;p2++) {
    if (Underlier_mask[p2]) ++count_g;
  }
  for(p2=0;p2<points;++p2){
    if( Underlier_mask[ Max1[t+1][p2] ] != 0 && Underlier_mask[ Max2[t+1][p2] ] != 0){
      S1  += Weight[p2];
      S2  += Weight[p2] * Weight[p2];
      S7  += Weight[p2] * Value[t+1][p2];
      S7U[ Max1[t+1][p2] ] += Weight[p2] * Value[t+1][p2];
      S8  += Weight[p2] * (g[p2] - G);
      S8U[ Max1[t+1][p2] ] += Weight[p2] * (g[p2] - G);
      S9  += Weight[p2] * Weight[p2] * Value[t+1][p2];
      S10 += Weight[p2] * Weight[p2] * (g[p2] - G);
      S12 += Weight[p2] * Weight[p2] * Value[t+1][p2] * (g[p2] - G);
      S13 += Weight[p2] * Weight[p2] * (g[p2] - G) * (g[p2] - G);    
    }
  }

  if (S1 ==0) printf("BUG3-->,%d\n",count_g);
  if((S13/S1 - 2*S8*S10/(S1*S1) + S2*S8*S8/(S1*S1*S1)) == 0){
    for(u=0;u<UNDERLIERS;++u){
      Prob[t][u][p] = S7U[u]/S1;
    }
    return(S7/S1);
  }
  
  beta = (S12/S1 - (S8*S9 + S7*S10)/(S1*S1) + S2*S7*S8/(S1*S1*S1))/
    (S13/S1 - 2*S8*S10/(S1*S1) + S2*S8*S8/(S1*S1*S1));


  /* DEBUG */
#if DEBUG
  for(p2=0;p2<points;++p2){
    printf("[%i][%i] -> [%i][:] Weight[%i] = %16g\t Value[%i][%i] = %f\tg[%i] = %f\n",t,p,t+1,p2,Weight[p2],t+1,p2,Value[t+1][p2],p2,g[p2]);
  }
  printf("Sum_weights = %g\n",Sum_weights);
  printf("beta = %g\n",beta);
  printf("(S1,S2,S7,S8,S9,S10,S12,S13,G,ge) = (%f,%f,%f,%f,%f,%f,%f,%f,%f,%f)\n",S1,S2,S7,S8,S9,S10,S12,S13,G,ge);
#endif

  if(fabs(beta) > 2.0){
    for(u=0;u<UNDERLIERS;++u){
      Prob[t][u][p] = S7U[u]/S1;
    }
     return(S7/S1);
  }

  for(u=0;u<UNDERLIERS;++u){
    Prob[t][u][p] = (S7U[u] - beta*S8U[u])/S1;
  }

  return((S7-beta*S8)/S1);
}


double mesh_estimate_0(int t, int p){
  int p2, points;
  double vih=0;

  weights1(t,p,t+1);

  points = points_on_timeslice(t+1);

  /* DEBUG */
#if DEBUG
  for(p2=0;p2<points;++p2){
    printf("[%i][%i] -> [%i][:] Weight[%i] = %16g\t Value[%i][%i] = %f\n",t,p,t+1,p2,Weight[p2],t+1,p2,Value[t+1][p2]);
  }
  printf("Sum_weights = %g\n",Sum_weights);
#endif


  for(p2=0;p2<points;++p2){
    vih += Weight[p2]*Value[t+1][p2];
  }
  return(vih/Sum_weights);
}
